/*

DRSSTC MIDI Interrupter
Eirik Taylor, 2013

Receives incoming MIDI signals with UART module, and plays corresponding note by
using Compare Match Modules to toggle OC1A and OC1B at the note's frequency. Uses
Compare Match A and B to get two note polyphony. Since only two notes can be played
at once per C, additional notes are ignored and played by other C further down.

An external Hex switch (4-bit output rotary switch, from 0 to 0xF) provides the channel
number.

PB1 - output - Note 1
PB2 - output - Note 2

PC0-3 - input - Channel selection

PD2 - input - Enable this unit (active low)
PD4 - output - Enable next unit
PD7 - output - Playing LED

Clock: 12.0 MHz
C: ATmega328

*/

// USART definitions
#define USART_BAUDRATE 31250
#define BAUD_PRESCALE (((F_CPU / (USART_BAUDRATE * 16UL))) - 1)

// Midi flags and key definitions
#define NOTE_ON 0
#define VALID 1
#define COUNTER 2
#define STATUS_BYTE 7
#define UPPERNOTE_INDEX 108
#define LOWERNOTE_INDEX 21

// Note definitions
#define OCTAVE_SIZE 12
#define OCTAVE_SHIFT 2
#define OCTAVE_INDEX (OCTAVE_SIZE*OCTAVE_SHIFT)
#define HIGHEST_NOTE 72

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "buffer.h"

volatile uint16_t compareB_off, compareA_off, compareB_on, compareA_on;
volatile uint8_t compareB_playing, compareA_playing, modeB, modeA;

// List of Timer 1 values required to generate various frequencies, encompasses full 88 notes.

static const uint16_t compare_off[88] PROGMEM = {54170, 51109, 48219, 45492, 42917, 40487, 38194, 36029, 33986, 32057, 30237, 28519, 26897, 25367, 23922, 22558, 21271, 20056, 18922, 17860, 16858, 15912, 15019, 14176, 13380, 12630, 11920, 11251, 10620, 10023, 9461, 8930, 8429, 7956, 7510, 7088, 6690, 6315, 5960, 5626, 5310, 5011, 4731, 4465, 4215, 3978, 3755, 3544, 3345, 3157, 2980, 2813, 2655, 2505, 2365, 2233, 2107, 1989, 1878, 1772, 1672, 1578, 1490, 1407, 1327, 1252, 1183, 1116, 1053, 994, 939, 886, 836, 789, 745, 703, 664, 626, 591, 558, 526, 497, 470, 443, 418, 395, 372, 352};

static const uint16_t compare_on[88] PROGMEM = {375, 375, 375, 375, 375, 375, 375, 375, 375, 375, 375, 375, 375, 375, 375, 375, 375, 375, 362, 342, 322, 304, 287, 271, 256, 241, 228, 215, 203, 192, 181, 171, 161, 152, 143, 135, 128, 120, 114, 107, 101, 96, 90, 85, 80, 76, 71, 67, 64, 60, 57, 53, 50, 48, 45, 42, 40, 38, 35, 33, 32, 30, 28, 26, 25, 24, 22, 21, 20, 19, 17, 16, 16, 15, 14, 13, 12, 12, 11, 10, 10, 9, 8, 8, 8, 7, 7, 6};

// Lookup Compare Match B value
void set_compB(uint8_t data) {
	compareB_playing = data;
	compareB_off = pgm_read_word(&(compare_off[data - LOWERNOTE_INDEX]));
	compareB_on = pgm_read_word(&(compare_on[data - LOWERNOTE_INDEX]));

	// Compare Match B interrupt enable
	modeB = 0;
	TIMSK1 |= (1 << OCIE1B);
}

// Lookup Compare Match B value
void set_compA(uint8_t data) {
	compareA_playing = data;
	compareA_off = pgm_read_word(&(compare_off[data - LOWERNOTE_INDEX]));
	compareA_on = pgm_read_word(&(compare_on[data - LOWERNOTE_INDEX]));

	// Compare Match A interrupt enable
	modeA = 0;
	TIMSK1 |= (1 << OCIE1A);
}


// Find who's playing the note and shut them up
void stop(uint8_t data) {
	// Shift key down two octaves...
	if (data >= OCTAVE_INDEX) {
		data -= OCTAVE_INDEX;
		
		// Check if playing notes are to be shut off
		if (compareB_playing == data) {
			TIMSK1  &=~(1 << OCIE1B);
			// Set note output low
			PORTB &=~(1<<PB2);
			// Set dual note indicator high
			PORTB |= (1<<PD4);
			compareB_playing = 0;
			
			}else if (compareA_playing == data) {
			TIMSK1  &=~(1 << OCIE1A);
			// Set note output low
			PORTB &=~(1<<PB1);
			// Set dual note indicator high
			PORTB |= (1<<PD4);
			compareA_playing = 0;
		}
	}
}

// Compare Match B interrupt
ISR(TIMER1_COMPB_vect) {
	// Toggle output pin
	PORTB ^= (1<<PB2);
	
	// add start value to current timer 1 value ensures that compare match occurrs at set number of Timer 1 increments
	if (modeB) {
		OCR1B = TCNT1 + compareB_off;
		modeB = 0;
		}else {
		OCR1B = TCNT1 + compareB_on;
		modeB = 1;
	}
}

// Compare Match A interrupt
ISR(TIMER1_COMPA_vect) {
	// Toggle output pin
	PORTB ^= (1<<PB1);

	// add start value to current timer 1 value ensures that compare match occurrs at set number of Timer 1 increments
	if (modeA) {
		OCR1A = TCNT1 + compareA_off;
		modeA = 0;
		}else {
		OCR1A = TCNT1 + compareA_on;
		modeA = 1;
	}
}

// Pick compare module to play note
void play(uint8_t data);
// Initialize USART Module
void USART_Init(void);
// USART Received Data Complete interrupt service routine
ISR(USART_RX_vect) {
	// Place the USART contents in the buffer, so it can be handled later
	bufferInsert(UDR0);
}

void handleMIDIdata(uint8_t data, uint8_t channel);
// Main program Loop
int main(void) {
	uint8_t midChannel;

	// PORTB Initialization, PB1 and PB2 are note outputs
	DDRB = (1<<PB2)|(1<<PB1);
	PORTB = 0;

	// PORTD Initialization, PD7 for "Playing LED" and PD4 for "next unit enable".
	DDRD = (1<<PD7)|(1<<PD4);
	// Need pull-up on enable input PD2, active low.
	PORTD = (1<<PD2);

	// Set MIDI channel to listen on
	DDRC = 0;
	// Channel selector is inverted cathod switch. Need pull-ups.
	PORTC = 0x0F;

	// Initialize USART module
	USART_Init();

	// Start Timer 1 with prescaler of 1:8
	TCCR1B |= (1 << CS11);

	// Initialize USART buffer
	writeIndex = 0;
	currentIndex = 0;

	// Enable global interrupts
	sei();
	
	while (1) {
		
		// Update Channel (problems if doing this outside of main loop, might have to do with start up voltage rise.
		midChannel = ~PINC;
		midChannel &= 0x0F;
		
		// Check for new data in the USART buffer
		if (writeIndex != currentIndex) {
			// Work on the data
			handleMIDIdata(popFromBuffer(), midChannel);
		}

		// Check if notes are being played
		if ((compareB_playing != 0) || (compareA_playing != 0)) {
			// Light status LED
			PORTD |= (1<<PD7);
			}else{
			// Turn off status LED
			PORTD &=~(1<<PD7);
		}
		
		// Check if next AVR in line should play or not
		if ((compareB_playing != 0) && (compareA_playing != 0) && ((~PIND) & (1<<PD2))) {
			// if true, both modules are now playing, and the previous AVR is full
			// Enable next AVR in chain, active low
			PORTD &=~(1<<PD4);
			
			}else{
			// Keep next AVR in standby
			PORTD |= (1<<PD4);
		}
	}
	
	return 0;
}

// Handle new MIDI data
void handleMIDIdata(uint8_t data, uint8_t channel) {
	uint8_t temp = data;
	static uint8_t midi_flags, current_note;
	
	// Determine if status, data, or velocity byte. Remember note on and off events. Ignore irrelevant data bytes.
	if (temp < 0b11111000) {

		// if Status byte
		if (temp & (1<<STATUS_BYTE)) {
			// Default to ignore following bytes
			midi_flags &=~(1 << VALID);
			
			// Set flags based on status byte type, currently only "note on" and "note off" are used, channel is chosen.
			if (temp == (0b10000000 | channel)) {
				// Note OFF event
				midi_flags |= (1 << VALID);
				midi_flags  &=~(1 << NOTE_ON);
				
				}else if (temp == (0b10010000 | channel)) {
				// Note ON event
				midi_flags |= (1 << NOTE_ON);
				midi_flags |= (1 << VALID);
			}

			// If data byte number 2 and status byte was valid
			}else if ((bit_is_set(midi_flags, COUNTER)) && (bit_is_set(midi_flags, VALID))) {
			// if counter is set, first data byte has already been received, current databyte is velocity and final byte in MIDI word
			
			// Clear counter
			midi_flags &=~(1 << COUNTER);
			
			// Check for alternate "Note OFF" event
			if (temp == 0) {
				midi_flags  &=~(1 << NOTE_ON);
			}
			
			// Process data since full MIDI word has been received
			// if Note OFF flag is set
			if (bit_is_clear(midi_flags, NOTE_ON)) {
				// Turn off Note
				stop(current_note);
				
				// Otherwise check if Note ON flag is set before playing a note
				}else if (bit_is_set(midi_flags, NOTE_ON)) {
				// Check if it's this C's turn to play notes
				if ((~PIND) & (1<<PD2)) {
					// Play Note!
					play(current_note);
				}
			}

			// Only other option is a data byte, check that the status byte was valid first
			}else if (bit_is_set(midi_flags, VALID)) {
			midi_flags |= (1 << COUNTER);
			current_note = temp;
		}
	}
}

// Initialize USART Module
void USART_Init(void) {
	//Enable Receiver and Interrupt on receive complete
	UCSR0B = (1<<RXCIE0)|(1<<RXEN0);
	//Set data frame format: asynchronous mode,no parity, 1 stop bit, 8 bit size
	UCSR0C = (1<<UCSZ01)|(1<<UCSZ00);

	// Load lower 8-bits of the baud rate value into the low byte of the UBRR register
	UBRR0L = BAUD_PRESCALE;
	// Load upper 8-bits of the baud rate value into the high byte of the UBRR register
	UBRR0H = (BAUD_PRESCALE >> 8);
}

// Pick compare module to play note
void play(uint8_t data) {
	// Shift key down two octaves...
	if (data >= OCTAVE_INDEX) {
		data -= OCTAVE_INDEX;
	}
	//...and check if in range
	if ((data <= HIGHEST_NOTE) && (data >= LOWERNOTE_INDEX )) {
		// See which compare module isn't busy
		if (compareB_playing == 0) {
			set_compB(data);
			}else if (compareA_playing == 0) {
			set_compA(data);
		}
	}
}
