/*

Tesla Coil MIDI Interrupter
Eirik Taylor, 2009

Receives incoming MIDI signals with UART module, and plays corresponding note by
using Compare Match Modules to toggle PD4 and PD5 at the note's frequency. Uses
Compare Match A and B to get two note polyphony. Since only two notes can be played
at once per C, additional notes are ignored and played by other C further down.

An external Hex switch (4-bit output rotary switch, from 0 to 0xF) provides the channel
number.

Clock: 12.0 MHz
C: ATtiny2313

*/

// USART definitions
#define USART_BAUDRATE 31250
#define BAUD_PRESCALE (((F_CPU / (USART_BAUDRATE * 16UL))) - 1)

// Midi flags definitions
#define NOTE_ON 0
#define VALID 1
#define COUNTER 2

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

volatile uint16_t compareB_off, compareA_off, compareB_on, compareA_on;
volatile uint8_t compareB_playing, compareA_playing, modeB, modeA, channel;

// List of Timer 1 values required to generate various frequencies, encompasses full 88 notes.

static const uint16_t compare_off[88] PROGMEM = {51818, 48910, 46165, 43574, 41128, 38819, 36641, 34584, 32643, 30811, 29082, 27450, 25909, 24455, 23083, 21787, 20564, 19410, 18320, 17292, 16321, 15406, 14541, 13725, 12955, 12228, 11541, 10893, 10282, 9705, 9160, 8646, 8161, 7703, 7271, 6862, 6478, 6114, 5771, 5447, 5141, 4852, 4580, 4323, 4081, 3852, 3635, 3431, 3239, 3057, 2886, 2723, 2570, 2426, 2290, 2162, 2040, 1926, 1818, 1715, 1619, 1528, 1443, 1362, 1285, 1213, 1145, 1081, 1020, 963, 909, 857, 810, 764, 722, 681, 643, 607, 572, 540, 510, 481, 455, 429, 405, 382, 361, 341};

static const uint16_t compare_on[88] PROGMEM = {2727, 2574, 2429, 2293, 2164, 2043, 1928, 1820, 1718, 1621, 1530, 1444, 1363, 1287, 1214, 1146, 1082, 1021, 964, 910, 859, 810, 765, 722, 681, 643, 607, 573, 541, 510, 482, 455, 429, 405, 382, 361, 340, 321, 303, 286, 270, 255, 241, 227, 214, 202, 191, 180, 170, 160, 151, 143, 135, 127, 120, 113, 107, 101, 95, 90, 85, 80, 75, 71, 67, 63, 60, 56, 53, 50, 47, 45, 42, 40, 37, 35, 33, 31, 30, 28, 26, 25, 23, 22, 21, 20, 18, 17};

// Lookup Compare Match B value
void set_compB(uint8_t data)
{
	compareB_playing = data;

	// Check if key is in range, if not, assign limiting value.
	if (data > 108)
	{
		compareB_off = pgm_read_word(&(compare_off[87]));
		compareB_on = pgm_read_word(&(compare_on[87]));
	}
	else if (data < 21)
	{
		compareB_off = pgm_read_word(&(compare_off[0]));
		compareB_on = pgm_read_word(&(compare_on[0]));
		// In range!
	}
	else
	{
		compareB_off = pgm_read_word(&(compare_off[data - 21]));
		compareB_on = pgm_read_word(&(compare_on[data - 21]));
	}

	// Compare Match B interrupt enable
	modeB = 0;
	TIMSK |= (1 << OCIE1B);
}

// Lookup Compare Match B value
void set_compA(uint8_t data)
{
	compareA_playing = data;

	// Check if key is in range, if not, assign limiting value.
	if (data > 108)
	{
		compareA_off = pgm_read_word(&(compare_off[87]));
		compareA_on = pgm_read_word(&(compare_on[87]));
	}
	else if (data < 21)
	{
		compareA_off = pgm_read_word(&(compare_off[0]));
		compareA_on = pgm_read_word(&(compare_on[0]));
		// In range!
	}
	else
	{
		compareA_off = pgm_read_word(&(compare_off[data - 21]));
		compareA_on = pgm_read_word(&(compare_on[data - 21]));
	}

	// Compare Match A interrupt enable
	modeA = 0;
	TIMSK |= (1 << OCIE1A);
}


// Find who's playing the note and shut them up
void stop(uint8_t data)
{
	// Check if playing notes are to be shut off
	if (compareB_playing == data)
	{
		TIMSK  &=~(1 << OCIE1B);
		// Set note output low
		PORTB &=~(1);
		// Set dual note indicator low
		PORTB &=~(4);
		compareB_playing = 0;
		
	}
	else if (compareA_playing == data)
	{
		TIMSK  &=~(1 << OCIE1A);
		// Set note output low
		PORTB &=~(2);
		// Set dual note indicator low
		PORTB &=~(4);
		compareA_playing = 0;
	}
}

// Pick compare module to play note
void play(uint8_t data)
{
	// See which compare module isn't busy
	if (compareB_playing == 0)
	{
		set_compB(data);
	}
	else if (compareA_playing == 0)
	{
		set_compA(data);
	}
}

// Compare Match B interrupt
ISR(TIMER1_COMPB_vect)
{
	// Toggle output pin
	PORTB ^= 1;

	// add start value to current timer 1 value ensures that compare match occurs at set number of Timer 1 increments
	if (modeB)
	{
		OCR1B = TCNT1 + compareB_off;
		modeB = 0;
	}
	else
	{
		OCR1B = TCNT1 + compareB_on;
		modeB = 1;
	}
}

// Compare Match A interrupt
ISR(TIMER1_COMPA_vect)
{
	// Toggle output pin
	PORTB ^= 2;
	
	// add start value to current timer 1 value ensures that compare match occurs at set number of Timer 1 increments
	if (modeA)
	{
		OCR1A = TCNT1 + compareA_off;
		modeA = 0;
	}
	else
	{
		OCR1A = TCNT1 + compareA_on;
		modeA = 1;
	}
}

// Initialize USART Module
void USART_Init(void)
{
	//Enable Receiver and Interrupt on receive complete
	UCSRB |= (1 << RXEN)|(1<<RXCIE);
	//Set data frame format: asynchronous mode,no parity, 1 stop bit, 8 bit size
	UCSRC = (0<<UMSEL0)|(0<<UPM1)|(0<<UPM0)|(0<<USBS)|(1<<UCSZ1)|(1<<UCSZ0);

	// Load lower 8-bits of the baud rate value into the low byte of the UBRR register
	UBRRL = BAUD_PRESCALE;
	// Load upper 8-bits of the baud rate value into the high byte of the UBRR register
	UBRRH = (BAUD_PRESCALE >> 8);
}

// USART Received Data Complete interrupt service routine
ISR(USART_RX_vect)
{
	uint8_t temp;
	static uint8_t midi_flags, current_note;
	
	// Store value from USART Buffer
	temp = UDR;
	
	// Determine if status, data, or velocity byte. Remember note on and off events. Ignore irrelevant data bytes.
	if (temp < 0b11111000)
	{
		// if Status byte
		if (temp & 0b10000000)
		{
			// Default to ignore following bytes
			midi_flags &=~(1 << VALID);
			
			// Set flags based on status byte type, currently only "note on" and "note off" are used, channel is set by switches at start-up.
			if (temp == (0b10000000 | channel))
			{
				// Note OFF event
				midi_flags |= (1 << VALID);
				midi_flags &=~(1 << NOTE_ON);
				
			}
			else if (temp == (0b10010000 | channel))
			{
				// Note ON event
				midi_flags |= (1 << NOTE_ON);
				midi_flags |= (1 << VALID);
			}

			// If data byte number 2 and status byte was valid
		}
		else if ((bit_is_set(midi_flags, COUNTER)) && (bit_is_set(midi_flags, VALID)))
		{
			// if counter is set, first data byte has already been received, current data-byte is velocity and final byte in MIDI word
			
			// Clear counter
			midi_flags &=~(1 << COUNTER);
			
			// Check for alternate "Note OFF" event
			if (temp == 0)
			{
				midi_flags  &=~(1 << NOTE_ON);
			}
			
			// Process data since full MIDI word has been received
			// if Note OFF flag is set
			if (bit_is_clear(midi_flags, NOTE_ON))
			{
				// Turn off Note
				stop(current_note);
				
				// Otherwise check if Note ON flag is set before playing a note
			}
			else if (bit_is_set(midi_flags, NOTE_ON))
			{
				// Check if it's this C's turn to play notes
				if (PIND & 0b0001000)
				{
					// Play Note!
					play(current_note);
				}
				
			}

			// Only other option is a data byte, check that the status byte was valid first
		}
		else if (bit_is_set(midi_flags, VALID))
		{
			midi_flags |= (1 << COUNTER);
			current_note = temp;
		}
	}
}

// Main program Loop
int main(void)
{
	// PORTB Initialization, set PB0-3 as output
	// PB0 and PB1 are note outputs, PB3 indicates when notes are being played and PB2 indicates when 2 notes are being played.
	PORTB = 0;
	DDRB = 0x0F;

	// PORTD Initialization, all inputs
	PORTD = 0;
	DDRD = 0;

	// Set MIDI channel to listen on
	channel = ((PINB & 0xF0) >> 4);

	// Initialize USART module
	USART_Init();

	// Start Timer 1 with pre-scaler of 1:8
	TCCR1B |= (1 << CS11);

	// Enable global interrupts
	sei();

	// eternal loop. as is evident, the core of the program is carried out by interrupts.
	while (1)
	{
		// Check if notes are being played
		if ((compareB_playing != 0) || (compareA_playing != 0))
		{
			// Light status LED
			PORTB |= (1 << PB3);
		}
		else
		{
			// Turn off status LED
			PORTB &=~(1 << PB3);
		}
		
		// Check if next AVR in line should play or not
		if ((compareB_playing != 0) && (compareA_playing != 0) && (PIND & 0b0001000))
		// if true, both modules are now playing, and the previous AVR is full
		{
			// Enable next AVR in chain
			PORTB |= (4);
		}
		else
		{
			// Keep next AVR in standby
			PORTB &=~(4);
		}
		
		// Check MIDI channel to listen on, again, just so AVR responds to switch changes without needing a restart
		channel = ((PINB & 0xF0) >> 4);
	}
	
	return 0;
}
